"use strict";
var __create = Object.create;
var __defProp = Object.defineProperty;
var __getOwnPropDesc = Object.getOwnPropertyDescriptor;
var __getOwnPropNames = Object.getOwnPropertyNames;
var __getProtoOf = Object.getPrototypeOf;
var __hasOwnProp = Object.prototype.hasOwnProperty;
var __copyProps = (to, from, except, desc) => {
  if (from && typeof from === "object" || typeof from === "function") {
    for (let key of __getOwnPropNames(from))
      if (!__hasOwnProp.call(to, key) && key !== except)
        __defProp(to, key, { get: () => from[key], enumerable: !(desc = __getOwnPropDesc(from, key)) || desc.enumerable });
  }
  return to;
};
var __toESM = (mod, isNodeMode, target) => (target = mod != null ? __create(__getProtoOf(mod)) : {}, __copyProps(
  // If the importer is in node compatibility mode or this is not an ESM
  // file that has been converted to a CommonJS file using a Babel-
  // compatible transform (i.e. "__esModule" has not been set), then set
  // "default" to the CommonJS "module.exports" for node compatibility.
  isNodeMode || !mod || !mod.__esModule ? __defProp(target, "default", { value: mod, enumerable: true }) : target,
  mod
));

// src/ServerlessLiveLambdaPlugin.ts
var import_client_lambda2 = require("@aws-sdk/client-lambda");
var import_dotenv = __toESM(require("dotenv"));
var import_dotenv_expand = __toESM(require("dotenv-expand"));
var import_promises4 = __toESM(require("fs/promises"));
var import_path7 = __toESM(require("path"));

// src/bus.ts
var import_crypto = require("crypto");

// src/utils/lazy.ts
function lazy(callback) {
  let loaded = false;
  let result;
  return () => {
    if (!loaded) {
      result = callback();
      loaded = true;
    }
    return result;
  };
}

// src/bus.ts
var useBus = lazy(() => {
  const subscriptions = {};
  function subscribers(type) {
    let arr = subscriptions[type];
    if (!arr) {
      arr = [];
      subscriptions[type] = arr;
    }
    return arr;
  }
  const sourceID = (0, import_crypto.randomBytes)(16).toString("hex");
  return {
    sourceID,
    publish(type, properties) {
      const payload = {
        type,
        properties,
        sourceID
      };
      for (const sub of subscribers(type)) {
        sub.cb(payload);
      }
    },
    unsubscribe(sub) {
      const arr = subscribers(sub.type);
      const index = arr.indexOf(sub);
      if (index < 0)
        return;
      arr.splice(index, 1);
    },
    subscribe(type, cb) {
      const sub = {
        type,
        cb
      };
      subscribers(type).push(sub);
      return sub;
    },
    forward(..._types) {
      return (type, cb) => {
        return this.subscribe(type, cb);
      };
    }
  };
});

// src/console.ts
var import_randomcolor2 = __toESM(require("randomcolor"));

// src/logger.ts
var import_chalk = __toESM(require("chalk"));
var import_randomcolor = __toESM(require("randomcolor"));

// src/serverless.ts
var sls;
var setServerless = (serverless) => {
  sls = serverless;
};
var useServerless = () => {
  return sls;
};
var serverlessLog;
var setLog = (log) => {
  serverlessLog = log;
};
var useLog = () => {
  return serverlessLog;
};
var useFunctions = lazy(() => {
  const functions = sls.service.functions;
  let filtered = functions;
  let regex = new RegExp(".*");
  return {
    get all() {
      return filtered;
    },
    applyFilter: (pattern) => {
      regex = new RegExp(pattern);
      filtered = Object.keys(functions).filter((func) => {
        return regex.test(func);
      }).reduce((prev, current) => {
        return {
          ...prev,
          [current]: functions[current]
        };
      }, {});
    },
    fromId: (id) => {
      return Object.values(filtered).find((func) => func.name === id);
    }
  };
});
var serverlessOpts;
var setServerlessOptions = (opts) => {
  serverlessOpts = opts;
};
var useServerlessOptions = () => {
  return serverlessOpts;
};
var debug = false;
var setDebugMode = () => {
  debug = true;
};
var isDebug = () => {
  return debug;
};

// src/utils/time.ts
function timeNow() {
  const now = /* @__PURE__ */ new Date();
  return `${now.toLocaleDateString()} ${now.toLocaleTimeString()}`;
}

// src/logger.ts
var useColorMap = lazy(() => {
  const colorMap = /* @__PURE__ */ new Map();
  const functions = useFunctions().all;
  for (const func of Object.values(functions)) {
    colorMap.set(func.name, (0, import_randomcolor.default)({ luminosity: "light" }));
  }
  return colorMap;
});
var useFunctionLog = (functionId) => {
  const log = useLog();
  const colorMap = useColorMap();
  const props = useFunctions().fromId(functionId);
  const color = import_chalk.default.hex(colorMap.get(functionId));
  return {
    debug: (message, ...args) => {
      log.debug(
        `${color(props?.name)} ${color(timeNow())} ${message}`,
        ...args
      );
    },
    info: (message, ...args) => {
      log.notice(
        `${color(props?.name)} ${color(timeNow())} ${message}`,
        ...args
      );
    },
    success: (message, ...args) => {
      log.success(
        `${color(props?.name)} ${color(timeNow())} ${message}`,
        ...args
      );
    },
    warning: (message, ...args) => {
      log.warning(
        `${color(props?.name)} ${color(timeNow())} ${message}`,
        ...args
      );
    },
    danger: (message, ...args) => {
      log.error(
        `${color(props?.name)} ${color(timeNow())} ${message}`,
        ...args
      );
    },
    bold: import_chalk.default.bold,
    all: import_chalk.default,
    hex: import_chalk.default.hex,
    primary: import_chalk.default.hex("#FF9000")
  };
};
var useGlobalLog = () => {
  const log = useLog();
  return {
    debug: (message, ...args) => {
      log.debug(`${message}`, ...args);
    },
    info: (message, ...args) => {
      log.notice(`${message}`, ...args);
    },
    success: (message, ...args) => {
      log.success(`${message}`, ...args);
    },
    warning: (message, ...args) => {
      log.warning(`${message}`, ...args);
    },
    danger: (message, ...args) => {
      log.error(`${message}`, ...args);
    },
    bold: import_chalk.default.bold,
    all: import_chalk.default,
    hex: import_chalk.default.hex,
    primary: import_chalk.default.hex("#FF9000")
  };
};

// src/console.ts
var useFunctionColors = lazy(() => {
  const colorMap = /* @__PURE__ */ new Map();
  const functions = useFunctions().all;
  for (const func of Object.values(functions)) {
    colorMap.set(func.name, (0, import_randomcolor2.default)({ luminosity: "light" }));
  }
  return colorMap;
});
var useConsole = lazy(() => {
  const bus = useBus();
  const metrics = /* @__PURE__ */ new Map();
  bus.subscribe("function.build.started", (evt) => {
    const functionId = evt.properties.functionId;
    const log = useGlobalLog();
    let metric = metrics.get(functionId);
    if (!metric) {
      metric = {};
      metrics.set(functionId, metric);
    }
    metric.startBuildTime = /* @__PURE__ */ new Date();
    log.success(`Rebuild ${functionId} lambda`);
  });
  bus.subscribe("function.build.success", (evt) => {
    const log = useGlobalLog();
    const metric = metrics.get(evt.properties.functionId);
    metric.stopBuildTime = /* @__PURE__ */ new Date();
    const takenTime = (metric.stopBuildTime.getTime() - metric.startBuildTime.getTime()) / 1e3;
    log.success(`Done in ${takenTime}s`);
  });
  bus.subscribe("function.build.failed", (evt) => {
    const log = useGlobalLog();
    log.danger(`${evt.properties.functionId} built failed`);
    log.info(evt.properties.errors.join("\n"));
  });
  bus.subscribe("worker.stdout", (evt) => {
    const log = useFunctionLog(evt.properties.functionId);
    log.info(evt.properties.message);
  });
});

// src/credentials.ts
var useAWSProvider = lazy(() => {
  const sls2 = useServerless();
  const provider = sls2.getProvider("aws");
  if (!provider) {
    throw new Error("serverless-live-lambda: only AWS is supported");
  }
  return provider;
});
var useAWSCredentials = lazy(async () => {
  const provider = useAWSProvider();
  const { credentials } = provider.getCredentials();
  return credentials;
});
var useClientCache = lazy(() => /* @__PURE__ */ new Map());
function useAWSClient(client, force = false) {
  const log = useLog();
  const cache = useClientCache();
  const existing = cache.get(client.name);
  if (existing && !force)
    return existing;
  const [provider, credentials] = [useAWSProvider(), useAWSCredentials()];
  const result = new client({
    region: provider.getRegion(),
    credentials
  });
  cache.set(client.name, result);
  log.info(`Created AWS client ${client.name}`);
  return result;
}

// src/iot.ts
var import_client_iot = require("@aws-sdk/client-iot");
var import_client_s3 = require("@aws-sdk/client-s3");
var import_aws_iot_device_sdk = __toESM(require("aws-iot-device-sdk"));
var import_crypto2 = require("crypto");
var import_path = __toESM(require("path"));

// src/errors.ts
var VisibleError = class extends Error {
  constructor(...message) {
    super(message.join("\n"));
  }
};

// src/utils/zip.ts
var import_node_zlib = __toESM(require("node:zlib"));
function gzip(input) {
  return new Promise((resolve, reject) => {
    import_node_zlib.default.gzip(input, (err, buf) => {
      if (err) {
        reject(err);
        return;
      }
      resolve(buf);
    });
  });
}

// src/iot.ts
var useIOTEndpoint = lazy(async () => {
  const log = useGlobalLog();
  const iot2 = useAWSClient(import_client_iot.IoTClient);
  log.debug("Getting IoT endpoint");
  const response = await iot2.send(
    new import_client_iot.DescribeEndpointCommand({
      endpointType: "iot:Data-ATS"
    })
  );
  log.debug(`Using IoT endpoint: ${response.endpointAddress}`);
  if (!response.endpointAddress) {
    throw new VisibleError("IoT Endpoint address not found");
  }
  return response.endpointAddress;
});
var useIOTClientId = lazy(() => {
  const clientId = (0, import_crypto2.randomUUID)();
  return clientId;
});
var useIOT = lazy(async () => {
  const log = useGlobalLog();
  const bus = useBus();
  const sls2 = useServerless();
  const endpoint = await useIOTEndpoint();
  const provider = useAWSProvider();
  const cred = await useAWSCredentials();
  const serviceName = sls2.service.getServiceName();
  const stage = provider.getStage();
  async function encode(input) {
    const s3 = useAWSClient(import_client_s3.S3Client);
    const id = (0, import_crypto2.randomUUID)();
    const json = JSON.stringify(input);
    if (json.length > 1024 * 1024) {
      const deployBucket = await provider.getServerlessDeploymentBucketName();
      const key = import_path.default.join("pointers", id);
      const compressed = await gzip(json);
      await s3.send(
        new import_client_s3.PutObjectCommand({
          Bucket: deployBucket,
          Key: key,
          Body: compressed
        })
      );
      return [
        {
          id,
          index: 0,
          count: 1,
          data: JSON.stringify({
            type: "pointer",
            properties: {
              bucket: deployBucket,
              key,
              gzip: true
            }
          })
        }
      ];
    }
    const parts = json.match(/.{1,50000}/g);
    if (!parts)
      return [];
    log.debug(`Encoded iot message into ${parts.length} parts`);
    return parts.map((part, index) => ({
      id,
      index,
      count: parts?.length,
      data: part
    }));
  }
  const clientId = useIOTClientId();
  const device = new import_aws_iot_device_sdk.default.device({
    protocol: "wss",
    host: endpoint,
    region: provider.getRegion(),
    clientId,
    accessKeyId: cred.accessKeyId,
    secretKey: cred.secretAccessKey,
    sessionToken: cred.sessionToken,
    reconnectPeriod: 1
  });
  const PREFIX = `serverless/${serviceName}/${stage}`;
  device.subscribe(`${PREFIX}/events`, { qos: 1 });
  log.debug(`Subscribe to ${PREFIX}/events topic`);
  const fragments = /* @__PURE__ */ new Map();
  device.on("connect", () => {
    log.debug("IoT connected");
  });
  device.on("error", (err) => {
    log.debug(`IoT error ${err}`);
  });
  device.on("close", () => {
    log.debug("IoT closed");
  });
  device.on("reconnect", () => {
    log.debug("IoT reconnected");
  });
  device.on("message", (_topic, buffer) => {
    const fragment = JSON.parse(buffer.toString());
    let pending = fragments.get(fragment.id);
    if (!pending) {
      pending = /* @__PURE__ */ new Map();
      fragments.set(fragment.id, pending);
    }
    pending.set(fragment.index, fragment);
    if (pending.size === fragment.count) {
      const data = [...pending.values()].sort((a, b) => a.index - b.index).map((item) => item.data).join("");
      fragments.delete(fragment.id);
      const evt = JSON.parse(data);
      bus.publish(evt.type, evt.properties);
    }
  });
  return {
    prefix: PREFIX,
    async publish(topic, type, properties) {
      const payload = {
        type,
        properties,
        sourceID: bus.sourceID
      };
      for (const fragment of await encode(payload)) {
        await new Promise((r) => {
          device.publish(
            topic,
            JSON.stringify(fragment),
            {
              qos: 1
            },
            () => {
              r();
            }
          );
        });
      }
      log.debug(`IOT Published ${topic} ${type}`);
    }
  };
});

// src/process.ts
var import_client_lambda = require("@aws-sdk/client-lambda");
var useProcess = lazy(() => {
  const functions = useFunctions().all;
  const log = useGlobalLog();
  process.on("SIGINT", () => {
    const processes = [];
    for (const opts of Object.values(functions)) {
      const cleanup = cleanUpLambda(opts.name);
      processes.push(cleanup);
    }
    Promise.all(processes).then(() => process.exit(0)).catch((err) => {
      log.danger(
        [
          `Clean up error ${err}`,
          "Please manually remove `SLS_LIVE_LAMBDA_ENABLED=true`."
        ].join("\n")
      );
      process.exit(1);
    });
  });
});
async function cleanUpLambda(functionId) {
  const log = useGlobalLog();
  try {
    const lambda = useAWSClient(import_client_lambda.LambdaClient);
    const input = new import_client_lambda.GetFunctionConfigurationCommand({
      FunctionName: functionId
    });
    const output = await lambda.send(input);
    const envs = output.Environment?.Variables || {};
    delete envs["SLS_LIVE_LAMBDA_ENABLED"];
    await lambda.send(
      new import_client_lambda.UpdateFunctionConfigurationCommand({
        FunctionName: functionId,
        Environment: {
          Variables: envs
        }
      })
    );
    log.success(`Clean up ${functionId} success`);
  } catch (err) {
    log.danger(
      [
        `Clean up ${functionId} error`,
        "Please manually remove `SLS_LIVE_LAMBDA_ENABLED=true`."
      ].join("\n")
    );
  }
}

// src/runtime/handlers.ts
var import_promises3 = __toESM(require("fs/promises"));
var import_path6 = __toESM(require("path"));

// src/utils/semaphore.ts
var Semaphore = class {
  constructor(maxLocks = 1) {
    this.queue = [];
    this.locked = 0;
    this.maxLocks = maxLocks;
  }
  lock() {
    return new Promise((resolve) => {
      const unlock = () => {
        this.locked--;
        const next = this.queue.shift();
        if (next) {
          this.locked++;
          next(unlock);
        }
      };
      if (this.locked < this.maxLocks) {
        this.locked++;
        resolve(unlock);
      } else {
        this.queue.push(unlock);
      }
    });
  }
};

// src/watcher.ts
var import_chokidar = __toESM(require("chokidar"));
var import_path2 = __toESM(require("path"));
var useWatcher = lazy(() => {
  const sls2 = useServerless();
  const bus = useBus();
  const watcher = import_chokidar.default.watch([sls2.serviceDir], {
    persistent: true,
    ignoreInitial: true,
    followSymlinks: false,
    disableGlobbing: false,
    ignored: [
      "**/node_modules/**",
      "**/.build/**",
      "**/.sst/**",
      "**/.git/**",
      "**/debug.log",
      "**/.serverless/**",
      "**/vendor/**"
    ],
    awaitWriteFinish: {
      pollInterval: 100,
      stabilityThreshold: 20
    }
  });
  watcher.on("change", (file) => {
    bus.publish("file.changed", {
      file,
      relative: import_path2.default.relative(sls2.serviceDir, file)
    });
  });
  return {
    subscribe: bus.forward("file.changed")
  };
});

// src/runtime/handlers/go.ts
var import_child_process2 = require("child_process");
var import_os = __toESM(require("os"));
var import_path4 = __toESM(require("path"));

// src/utils/fs.ts
var import_promises = __toESM(require("fs/promises"));
var import_path3 = __toESM(require("path"));
async function findAbove(dir, target) {
  if (dir === "/")
    return void 0;
  if (await existsAsync(import_path3.default.join(dir, target)))
    return dir;
  return findAbove(import_path3.default.resolve(import_path3.default.join(dir, "..")), target);
}
function isChild(parent, child) {
  const relative = import_path3.default.relative(parent, child);
  return Boolean(
    relative && !relative.startsWith("..") && !import_path3.default.isAbsolute(relative)
  );
}
async function existsAsync(input) {
  return import_promises.default.access(input).then(() => true).catch(() => false);
}

// src/utils/process.ts
var import_child_process = require("child_process");
var import_util = require("util");
var execAsync = (0, import_util.promisify)(import_child_process.exec);

// src/runtime/server.ts
var import_express = __toESM(require("express"));
var import_https = __toESM(require("https"));
var import_url = require("url");

// src/runtime/workers.ts
var useRuntimeWorkers = lazy(async () => {
  const workers = /* @__PURE__ */ new Map();
  const bus = useBus();
  const builder = useFunctionBuilder();
  const handlers = useRuntimeHandlers();
  const server = await useRuntimeServerConfig();
  handlers.subscribe("function.build.success", async (evt) => {
    for (const [, worker] of workers) {
      if (worker.functionId === evt.properties.functionId) {
        const props = useFunctions().fromId(worker.functionId);
        if (!props) {
          return;
        }
        const handler = handlers.for(props.runtime);
        await handler?.stopWorker(worker.workerId);
        bus.publish("worker.stopped", worker);
      }
    }
  });
  const lastRequestId = /* @__PURE__ */ new Map();
  bus.subscribe("function.invoked", async (evt) => {
    bus.publish("function.ack", {
      functionId: evt.properties.functionId,
      workerId: evt.properties.workerId
    });
    lastRequestId.set(evt.properties.workerId, evt.properties.requestId);
    if (isDebug()) {
      workers.set(evt.properties.workerId, {
        workerId: evt.properties.workerId,
        functionId: evt.properties.functionId
      });
      return;
    }
    const worker = workers.get(evt.properties.workerId);
    if (worker)
      return;
    const props = useFunctions().fromId(evt.properties.functionId);
    if (!props)
      return;
    const handler = handlers.for(props.runtime);
    if (!handler)
      return;
    const build = await builder.artifact(evt.properties.functionId);
    if (!build)
      return;
    await handler.startWorker({
      ...build,
      workerId: evt.properties.workerId,
      functionId: evt.properties.functionId,
      environment: evt.properties.env,
      url: `${server.url}/${evt.properties.workerId}/${server.API_VERSION}`,
      runtime: props.runtime
    });
    workers.set(evt.properties.workerId, {
      workerId: evt.properties.workerId,
      functionId: evt.properties.functionId
    });
    bus.publish("worker.started", {
      workerId: evt.properties.workerId,
      functionId: evt.properties.functionId
    });
  });
  return {
    fromId(workerId) {
      return workers.get(workerId);
    },
    getCurrentRequestId(workerId) {
      return lastRequestId.get(workerId);
    },
    stdout(workerId, message) {
      const worker = workers.get(workerId);
      bus.publish("worker.stdout", {
        ...worker,
        message: message.trim(),
        requestId: lastRequestId.get(workerId)
      });
    },
    exited(workerId) {
      const existing = workers.get(workerId);
      if (!existing)
        return;
      workers.delete(workerId);
      lastRequestId.delete(workerId);
      bus.publish("worker.exited", existing);
    },
    subscribe: bus.forward(
      "worker.started",
      "worker.stopped",
      "worker.exited",
      "worker.stdout"
    )
  };
});

// src/runtime/server.ts
var useRuntimeServerConfig = lazy(async () => {
  const opts = useServerlessOptions();
  const port = parseInt(opts.port || opts.p || "18080");
  return {
    API_VERSION: "2018-06-01",
    port,
    url: `http://localhost:${port}`
  };
});
var useRuntimeServer = lazy(async () => {
  const log = useGlobalLog();
  const bus = useBus();
  const app = (0, import_express.default)();
  const workers = await useRuntimeWorkers();
  const cfg = await useRuntimeServerConfig();
  const workersWaiting = /* @__PURE__ */ new Map();
  const invocationsQueued = /* @__PURE__ */ new Map();
  function next(workerID) {
    const queue = invocationsQueued.get(workerID);
    const value = queue?.shift();
    if (value)
      return value;
    return new Promise((resolve) => {
      workersWaiting.set(workerID, resolve);
    });
  }
  workers.subscribe("worker.exited", async (evt) => {
    const waiting = workersWaiting.get(evt.properties.workerId);
    if (!waiting)
      return;
    workersWaiting.delete(evt.properties.workerId);
  });
  bus.subscribe("function.invoked", async (evt) => {
    const worker = workersWaiting.get(evt.properties.workerId);
    if (worker) {
      workersWaiting.delete(evt.properties.workerId);
      worker(evt.properties);
      return;
    }
    let arr = invocationsQueued.get(evt.properties.workerId);
    if (!arr) {
      arr = [];
      invocationsQueued.set(evt.properties.workerId, arr);
    }
    arr.push(evt.properties);
  });
  app.post(
    `/:workerId/${cfg.API_VERSION}/runtime/init/error`,
    import_express.default.json({
      strict: false,
      type: ["application/json", "application/*+json"],
      limit: "10mb"
    }),
    async (req, res) => {
      const worker = workers.fromId(req.params.workerId);
      bus.publish("function.error", {
        requestID: workers.getCurrentRequestId(worker.workerId),
        workerID: worker.workerId,
        functionID: worker.functionId,
        ...req.body
      });
      res.json("ok");
    }
  );
  app.get(
    `/:workerId/${cfg.API_VERSION}/runtime/invocation/next`,
    async (req, res) => {
      log.debug(
        "Worker",
        req.params.workerId,
        "is waiting for next invocation"
      );
      const payload = await next(req.params.workerId);
      log.debug("Worker", req.params.workerId, "sending next payload");
      res.set({
        "Lambda-Runtime-Aws-Request-Id": payload.requestId,
        "Lambda-Runtime-Deadline-Ms": Date.now() + payload.deadline,
        "Lambda-Runtime-Invoked-Function-Arn": payload.context.invokedFunctionArn,
        "Lambda-Runtime-Client-Context": JSON.stringify(
          payload.context.clientContext || null
        ),
        "Lambda-Runtime-Cognito-Identity": JSON.stringify(
          payload.context.identity || null
        ),
        "Lambda-Runtime-Log-Group-Name": payload.context.logGroupName,
        "Lambda-Runtime-Log-Stream-Name": payload.context.logStreamName
      });
      res.json(payload.event);
    }
  );
  app.post(
    `/:workerId/${cfg.API_VERSION}/runtime/invocation/:awsRequestId/response`,
    import_express.default.json({
      strict: false,
      type() {
        return true;
      },
      limit: "10mb"
    }),
    (req, res) => {
      log.debug("Worker", req.params.workerId, "got response");
      const worker = workers.fromId(req.params.workerId);
      bus.publish("function.success", {
        workerId: worker.workerId,
        functionId: worker.functionId,
        requestId: req.params.awsRequestId,
        body: req.body
      });
      res.status(202).send();
    }
  );
  app.all(
    "/proxy*",
    import_express.default.raw({
      type: "*/*",
      limit: "1024mb"
    }),
    (req, res) => {
      res.header("Access-Control-Allow-Origin", "*");
      res.header(
        "Access-Control-Allow-Methods",
        "GET, PUT, PATCH, POST, DELETE"
      );
      res.header(
        "Access-Control-Allow-Headers",
        req.header("access-control-request-headers")
      );
      if (req.method === "OPTIONS")
        return res.send();
      const u = new import_url.URL(req.url.substring(7));
      const forward = import_https.default.request(
        u,
        {
          headers: {
            ...req.headers,
            host: u.hostname
          },
          method: req.method
        },
        (proxied) => {
          res.status(proxied.statusCode);
          for (const [key, value] of Object.entries(proxied.headers)) {
            res.header(key, value);
          }
          proxied.pipe(res);
        }
      );
      if (req.method !== "GET" && req.method !== "DELETE" && req.method !== "HEAD" && req.body) {
        forward.write(req.body);
      }
      forward.end();
      forward.on("error", (e) => {
        log.info(e.message);
      });
      return;
    }
  );
  app.post(
    `/:workerId/${cfg.API_VERSION}/runtime/invocation/:awsRequestId/error`,
    import_express.default.json({
      strict: false,
      type: ["application/json", "application/*+json"],
      limit: "10mb"
    }),
    (req, res) => {
      const worker = workers.fromId(req.params.workerId);
      bus.publish("function.error", {
        workerId: worker.workerId,
        functionId: worker.functionId,
        errorType: req.body.errorType,
        errorMessage: req.body.errorMessage,
        requestId: req.params.awsRequestId,
        trace: req.body.trace
      });
      res.status(202).send();
    }
  );
  log.success(`\u{1F680} API Runtime server listen on ${cfg.url}`);
  app.listen(cfg.port);
});

// src/runtime/handlers/go.ts
var useGoHandler = () => {
  const sls2 = useServerless();
  const processes = /* @__PURE__ */ new Map();
  const sources = /* @__PURE__ */ new Map();
  const handleName = import_os.default.platform() === "win32" ? "bootstrap.exe" : "bootstrap";
  return {
    shouldBuild: (input) => {
      const parent = sources.get(input.functionId);
      if (!parent) {
        return false;
      }
      return isChild(parent, input.file);
    },
    canHandle: (input) => input.startsWith("go") || input.startsWith("provided"),
    startWorker: async (input) => {
      const workers = await useRuntimeWorkers();
      const server = await useRuntimeServerConfig();
      const proc = (0, import_child_process2.spawn)(import_path4.default.join(input.out, handleName), {
        env: {
          ...process.env,
          ...input.environment,
          IS_LOCAL: "true",
          AWS_LAMBDA_RUNTIME_API: `localhost:${server.port}/${input.workerId}`
        },
        cwd: input.out
      });
      proc.on("exit", () => workers.exited(input.workerId));
      proc.stdout.on("data", (data) => {
        workers.stdout(input.workerId, data.toString());
      });
      proc.stderr.on("data", (data) => {
        workers.stdout(input.workerId, data.toString());
      });
      processes.set(input.workerId, proc);
    },
    stopWorker: async (workerId) => {
      const proc = processes.get(workerId);
      if (proc) {
        proc.kill();
        processes.delete(workerId);
      }
    },
    build: async (input) => {
      const props = input.props;
      const parsed = import_path4.default.parse(props.handler);
      const project = await findAbove(parsed.dir || ".", "go.mod");
      if (!project) {
        return {
          type: "error",
          errors: ["go.mod not found in current directory"]
        };
      }
      sources.set(input.functionId, project);
      try {
        const src = import_path4.default.relative(project, props.handler);
        const target = import_path4.default.join(input.out, handleName);
        const srcPath = import_os.default.platform() === "win32" ? src.replaceAll("\\", "\\\\") : src;
        let buildCmd = [
          "go",
          "build",
          '-ldflags "-s -w"',
          `-o "${target}"`,
          `./${srcPath}`
        ].join(" ");
        const custom = sls2.service.custom.go;
        if (custom?.cmd) {
          const cmd = custom.cmd;
          buildCmd = [
            cmd.replaceAll(/GOOS=(\S+)|GOARCH=(\S+)/gi, ""),
            `-o "${target}"`,
            `./${srcPath}`
          ].join(" ");
        }
        await execAsync(buildCmd, {
          cwd: project,
          env: {
            ...process.env
          }
        });
        return {
          type: "success",
          handler: "bootstrap"
        };
      } catch (err) {
        return {
          type: "error",
          errors: [String(err)]
        };
      }
    }
  };
};

// src/runtime/handlers/nodejs.ts
var import_esbuild = __toESM(require("esbuild"));
var import_fs2 = __toESM(require("fs"));
var import_promises2 = __toESM(require("fs/promises"));
var import_path5 = __toESM(require("path"));
var import_worker_threads = require("worker_threads");
var useNodeJsHandler = () => {
  const rebuildCache = {};
  const threads = /* @__PURE__ */ new Map();
  return {
    shouldBuild: (input) => {
      const cache = rebuildCache[input.functionId];
      if (!cache)
        return false;
      if (cache.shouldReload)
        return true;
      const relative = import_path5.default.relative(cache.project, input.file).split(import_path5.default.sep).join(import_path5.default.posix.sep);
      return Boolean(cache.result?.metafile?.inputs[relative]);
    },
    canHandle: (input) => input.startsWith("nodejs"),
    startWorker: async (input) => {
      const workers = await useRuntimeWorkers();
      const server = await useRuntimeServerConfig();
      const cache = rebuildCache[input.functionId];
      const nodeRic = cache.isESM ? "./node-ric/index.mjs" : "./node-ric/index.cjs";
      new Promise(() => {
        console.log("Start worker");
        const worker = new import_worker_threads.Worker(import_path5.default.join(__dirname, nodeRic), {
          env: {
            ...process.env,
            ...input.environment,
            IS_LOCAL: "true",
            AWS_LAMBDA_RUNTIME_API: `localhost:${server.port}/${input.workerId}`
          },
          execArgv: ["--enable-source-maps"],
          workerData: {
            ...input,
            out: cache.out,
            file: cache.file
          },
          stdout: true,
          stdin: true,
          stderr: true
        });
        worker.stdout.on("data", (data) => {
          workers.stdout(input.workerId, data.toString());
        });
        worker.stderr.on("data", (data) => {
          workers.stdout(input.workerId, data.toString());
        });
        worker.on("exit", () => {
          workers.exited(input.workerId);
        });
        threads.set(input.workerId, worker);
      });
    },
    stopWorker: async (workerId) => {
      const worker = threads.get(workerId);
      await worker?.terminate();
    },
    build: async (input) => {
      const [rootPath, exportFunction] = resolveHandler(input.props.handler);
      const extensions = [
        ".ts",
        ".tsx",
        ".mts",
        ".cts",
        ".js",
        ".jsx",
        ".mjs",
        ".cjs"
      ];
      const file = [
        rootPath,
        ...extensions.map((ext) => rootPath + ext),
        ...extensions.map((ext) => import_path5.default.join(rootPath, "index" + ext))
      ].find((file2) => {
        return import_fs2.default.existsSync(file2);
      });
      if (!file) {
        return {
          type: "error",
          errors: [`Could not find file for handler "${input.props.handler}"`]
        };
      }
      const project = await findAbove(rootPath || ".", "package.json");
      if (!project) {
        return {
          type: "error",
          errors: [
            `Could not find file package.json file for handler "${input.props.handler}"`
          ]
        };
      }
      const packageJson = JSON.parse(
        await import_promises2.default.readFile(import_path5.default.join(project, "package.json"), {
          encoding: "utf-8"
        })
      );
      const isESM = (packageJson.type || "") === "module";
      const canBuild = [".ts", ".tsx", ".mts", ".cts"].find(
        (ext) => file.endsWith(ext)
      );
      if (!canBuild) {
        rebuildCache[input.functionId] = {
          project,
          out: project,
          handler: exportFunction,
          file,
          shouldReload: true,
          isESM
        };
        return {
          type: "success",
          handler: exportFunction
        };
      }
      let ctx = rebuildCache[input.functionId]?.ctx;
      const outdir = import_path5.default.join(input.out, rootPath);
      if (!ctx) {
        const options = {
          entryPoints: [file],
          platform: "node",
          packages: "external",
          bundle: true,
          keepNames: true,
          logLevel: "silent",
          metafile: true,
          ...isESM ? {
            format: "esm",
            target: "esnext",
            mainFields: ["module", "main"],
            banner: {
              js: [
                "import { createRequire as topLevelCreateRequire } from 'module';",
                "const require = topLevelCreateRequire(import.meta.url);",
                'import { fileURLToPath as topLevelFileUrlToPath, URL as topLevelURL } from "url"',
                'const __dirname = topLevelFileUrlToPath(new topLevelURL(".", import.meta.url))'
              ].join("\n")
            }
          } : {
            format: "cjs",
            target: "node16"
          },
          outdir,
          sourcemap: true
        };
        ctx = await import_esbuild.default.context(options);
      }
      try {
        const result = await ctx.rebuild();
        rebuildCache[input.functionId] = {
          project,
          out: outdir,
          handler: exportFunction,
          file,
          shouldReload: false,
          ctx,
          result,
          isESM
        };
        return {
          type: "success",
          handler: input.props.handler
        };
      } catch (ex) {
        const result = ex;
        if ("errors" in result) {
          return {
            type: "error",
            errors: result.errors.flatMap((x) => [
              x.text,
              x.location?.file || "",
              x.location?.line + "\u2502" + x.location?.lineText
            ])
          };
        }
        return {
          type: "error",
          errors: [ex.toString()]
        };
      }
    }
  };
};
function resolveHandler(handler) {
  const parts = handler.split(import_path5.default.sep);
  const [index, exportFunction] = parts[parts.length - 1].split(".");
  const rootPath = import_path5.default.join(...parts.slice(0, parts.length - 1), index);
  return [rootPath, exportFunction];
}

// src/runtime/handlers.ts
var useRuntimeHandlers = lazy(() => {
  const handlers = [useGoHandler(), useNodeJsHandler()];
  const sls2 = useServerless();
  const bus = useBus();
  const pendingBuilds = /* @__PURE__ */ new Map();
  const result = {
    subscribe: bus.forward("function.build.success", "function.build.failed"),
    register: (handler) => {
      handlers.push(handler);
    },
    for: (runtime) => {
      const result2 = handlers.find((x) => x.canHandle(runtime));
      if (!result2) {
        throw new Error(`${runtime} runtime is unsupported`);
      }
      return result2;
    },
    async build(functionId) {
      async function task() {
        const func = useFunctions().fromId(
          functionId
        );
        if (!func) {
          return {
            type: "error",
            errors: [`Function with ID "${functionId}" not found`]
          };
        }
        const runtime = func.runtime || sls2.service.provider.runtime || "unknown";
        const handler = result.for(runtime);
        const serverlessDir = import_path6.default.join(sls2.serviceDir, ".serverless");
        const out = import_path6.default.join(serverlessDir, "artifacts", functionId);
        await import_promises3.default.rm(out, { recursive: true, force: true });
        await import_promises3.default.mkdir(out, { recursive: true });
        bus.publish("function.build.started", { functionId });
        const built = await handler.build({
          functionId,
          out,
          props: func
        });
        if (built.type === "error") {
          bus.publish("function.build.failed", {
            functionId,
            errors: built.errors
          });
          return built;
        }
        bus.publish("function.build.success", { functionId });
        return {
          ...built,
          out
        };
      }
      const log = useGlobalLog();
      if (pendingBuilds.has(functionId)) {
        log.debug("Waiting on pending build", functionId);
        return pendingBuilds.get(functionId);
      }
      const promise = task();
      pendingBuilds.set(functionId, promise);
      log.debug("Building function", functionId);
      try {
        return await promise;
      } catch (err) {
        return {
          type: "error",
          errors: [String(err)]
        };
      } finally {
        pendingBuilds.delete(functionId);
      }
    }
  };
  return result;
});
var useFunctionBuilder = lazy(() => {
  const log = useGlobalLog();
  const debug2 = isDebug();
  const artifacts = /* @__PURE__ */ new Map();
  const handlers = useRuntimeHandlers();
  const semaphore = new Semaphore(4);
  const result = {
    artifact: (functionId) => {
      if (artifacts.has(functionId)) {
        return artifacts.get(functionId);
      }
      return result.build(functionId);
    },
    build: async (functionId) => {
      const unlock = await semaphore.lock();
      try {
        const result2 = await handlers.build(functionId);
        if (!result2)
          return;
        if (result2.type === "error")
          return;
        artifacts.set(functionId, result2);
        return artifacts.get(functionId);
      } finally {
        unlock();
      }
    }
  };
  const watcher = useWatcher();
  watcher.subscribe("file.changed", async (evt) => {
    if (!debug2) {
      const functions = useFunctions();
      for (const func of Object.values(functions.all)) {
        const handler = handlers.for(func.runtime);
        if (!handler.shouldBuild({
          functionId: func.name,
          file: evt.properties.file
        })) {
          continue;
        }
        await result.build(func.name);
        log.debug("Rebuilt function", func.name);
      }
    }
  });
  return result;
});

// src/runtime/iot.ts
var useIOTBridge = lazy(async () => {
  const bus = useBus();
  const iot2 = await useIOT();
  const topic = `${iot2.prefix}/events`;
  const debug2 = isDebug();
  const cache = /* @__PURE__ */ new Map();
  if (debug2) {
    bus.subscribe("function.invoked", (evt) => {
      cache.set(evt.properties.requestId, evt.properties.workerId);
      evt.properties.workerId = evt.properties.functionId;
    });
  }
  bus.subscribe("function.success", async (evt) => {
    if (debug2) {
      evt.properties.workerId = cache.get(evt.properties.requestId);
      cache.delete(evt.properties.requestId);
    }
    await iot2.publish(
      topic + "/" + evt.properties.workerId,
      "function.success",
      evt.properties
    );
  });
  bus.subscribe("function.error", async (evt) => {
    if (debug2) {
      evt.properties.workerId = cache.get(evt.properties.requestId);
      cache.delete(evt.properties.requestId);
    }
    await iot2.publish(
      topic + "/" + evt.properties.workerId,
      "function.error",
      evt.properties
    );
  });
  bus.subscribe("function.ack", async (evt) => {
    await iot2.publish(
      topic + "/" + evt.properties.workerId,
      "function.ack",
      evt.properties
    );
  });
});

// src/ServerlessLiveLambdaPlugin.ts
var ServerlessLiveLambdaPlugin = class {
  constructor(serverless, options, { log }) {
    this.hooks = {
      "start:run": this.run.bind(this)
    };
    this.commands = {
      start: {
        usage: "Start your lambda function locally",
        lifecycleEvents: ["run"],
        options: {
          port: {
            usage: 'Specify API Runtime Server port (e.g. "--port 18080"). Default port is 18080',
            shortcut: "p",
            type: "number"
          },
          mode: {
            usage: "Specify the program mode. Only the value `debug` is allowed.",
            shortcut: "m",
            type: "string"
          },
          filter: {
            usage: "Utilize a regex pattern to filter the functions. The default is all functions.",
            shortcut: "f",
            type: "string"
          }
        }
      }
    };
    this.serverless = serverless;
    this.serverless.service.provider.environment ||= {};
    this.serviceName = serverless.service.getServiceName();
    this.stage = serverless.service.provider.stage;
    this.debug = (options?.mode || options?.m) === "debug";
    this.filterOpt = options?.filter || options.f || ".*";
    this.loadEnvs();
    if (this.debug) {
      setDebugMode();
    }
    setServerless(serverless);
    setLog(log);
    setServerlessOptions(options);
  }
  async run() {
    const log = useGlobalLog();
    await useIOTBridge();
    log.warning(
      [
        "Automatically set SLS_LIVE_LAMBDA_ENABLED=true to forward messages from Lambda to the local machine.",
        "This option poses a potential danger in a production environment,",
        "so please use it exclusively for development purposes."
      ].join("\n")
    );
    useFunctions().applyFilter(this.filterOpt);
    const functions = useFunctions().all;
    if (this.debug) {
      const cfg = await useRuntimeServerConfig();
      for (const [functionId, opts] of Object.entries(functions)) {
        const envs = {
          ...process.env,
          ...opts.environment,
          IS_LOCAL: "true",
          AWS_LAMBDA_RUNTIME_API: `localhost:${cfg.port}/${opts.name}`
        };
        const envFile = import_path7.default.join(".serverless", `.env.${functionId}`);
        await import_promises4.default.writeFile(
          import_path7.default.join(process.cwd(), envFile),
          Object.keys(envs).map((key) => `${key}=${envs[key] || ""}`).join("\n")
        );
        log.success(`Write ${envFile} env file success.`);
      }
      log.info(
        [
          "Debug mode for the server started successfully.",
          "Now, there's only one step left to debug your code on your local machine.",
          "Configure your debugger to load the generated environment file located at `${workspaceFolder}/.serverless/.env.${function}`."
        ].join("\n")
      );
    }
    const promises = Promise.all([
      useProcess(),
      useConsole(),
      useBus(),
      useIOT(),
      useRuntimeWorkers(),
      useRuntimeServer()
    ]);
    await promises;
    const builder = useFunctionBuilder();
    for (const opts of Object.values(functions)) {
      await builder.build(opts.name);
      await this.updateLiveLambdaModeFunctionEnvs(opts.name);
    }
  }
  async getLambdaFunctionConfiguration(functionName) {
    const lambda = useAWSClient(import_client_lambda2.LambdaClient);
    const input = new import_client_lambda2.GetFunctionConfigurationCommand({
      FunctionName: functionName
    });
    return await lambda.send(input);
  }
  async updateLiveLambdaModeFunctionEnvs(functionId) {
    const log = useGlobalLog();
    let envs;
    try {
      const props = await this.getLambdaFunctionConfiguration(functionId);
      envs = props.Environment?.Variables;
    } catch (err) {
      log.danger(
        [
          `Get lambda function ${functionId} configuration error.`,
          "The error may be due to a lack of permission to access the lambda function,",
          "or the function may not be deployed. Ensure that the function is deployed before starting local development."
        ].join("\n")
      );
      throw err;
    }
    const lambda = useAWSClient(import_client_lambda2.LambdaClient);
    await lambda.send(
      new import_client_lambda2.UpdateFunctionConfigurationCommand({
        FunctionName: functionId,
        Environment: {
          Variables: {
            ...envs,
            SLS_SERVICE_NAME: this.serviceName,
            SLS_STAGE: this.stage,
            SLS_LIVE_LAMBDA_ENABLED: "true"
          }
        }
      })
    );
  }
  loadEnvs() {
    const envFiles = [".env", `.env.${this.stage}`, `.env.${this.stage}.local`];
    envFiles.map((fileName) => {
      const parsed = import_dotenv.default.config({ path: fileName });
      return import_dotenv_expand.default.expand(parsed).parsed;
    });
    const envs = {
      SLS_SERVICE_NAME: this.serverless.service.getServiceName(),
      SLS_STAGE: this.serverless.service.provider.stage
    };
    Object.entries(envs).forEach(([key, value]) => {
      this.serverless.service.provider.environment[key] = value;
    });
  }
};
var ServerlessLiveLambdaPlugin_default = ServerlessLiveLambdaPlugin;

// src/index.ts
module.exports = ServerlessLiveLambdaPlugin_default;
//# sourceMappingURL=index.js.map
